import dataclasses
import typing

from . import _getlib
from . import contributions as _contributions
from . import duration as _duration
from . import illustrators as _illustrators
from . import indicators as _indicators
from . import score as _score


@dataclasses.dataclass(frozen=True, slots=True, unsafe_hash=True)
class MetricModulation:
    r"""
    Metric modulation.

    ..  container:: example

        With notes:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'8."),
        ...     right_rhythm=abjad.Note("c'4."),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation #3 #1 #2 #1 #'(1 . 1)

    ..  container:: example

        With tuplets:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ...     right_rhythm=abjad.Note("c'4"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-lhs #2 #0 #4 #5 #2 #0 #'(1 . 1)

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=abjad.Tuplet((4, 5), "c'4"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-rhs #2 #0 #2 #0 #4 #5 #'(1 . 1)

    ..  container:: example

        With tuplets again:

        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c16."),
        ...     right_rhythm=abjad.Tuplet((2, 3), "c8"),
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup \abjad-metric-modulation-tuplet-rhs #4 #1 #3 #0 #2 #3 #'(1 . 1)

    ..  container:: example

        With ties:

        >>> notes = abjad.makers.make_notes([0], [(5, 16)])
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=notes,
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup { \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                }
            =
            \hspace #-0.5
            \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                            ~
                            c'16
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                } }

    ..  container:: example

        With ties and tuplets:

        >>> notes = abjad.makers.make_notes([0], [(5, 16)])
        >>> tuplet = abjad.Tuplet((2, 3), notes)
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c'4"),
        ...     right_rhythm=tuplet,
        ... )

        >>> lilypond_file = abjad.LilyPondFile(
        ...     [r'\include "abjad.ily"', metric_modulation]
        ... )
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(metric_modulation)
            >>> print(string)
            \markup { \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            c'4
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                }
            =
            \hspace #-0.5
            \score
                {
                    \context Score = "Score"
                    \with
                    {
                        \override SpacingSpanner.spacing-increment = 0.5
                        proportionalNotationDuration = ##f
                    }
                    <<
                        \context RhythmicStaff = "Rhythmic_Staff"
                        \with
                        {
                            \remove Time_signature_engraver
                            \remove Staff_symbol_engraver
                            \override Stem.direction = #up
                            \override Stem.length = 5
                            \override TupletBracket.bracket-visibility = ##t
                            \override TupletBracket.direction = #up
                            \override TupletBracket.minimum-length = 4
                            \override TupletBracket.padding = 1.25
                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                            \override TupletBracket.springs-and-rods = #ly:spanner::set-spacing-rods
                            \override TupletNumber.font-size = 0
                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                            tupletFullLength = ##t
                        }
                        {
                            \tweak edge-height #'(0.7 . 0)
                            \tuplet 3/2
                            {
                                c'4
                                ~
                                c'16
                            }
                        }
                    >>
                    \layout
                    {
                        indent = 0
                        ragged-right = ##t
                    }
                } }


    ..  container:: example

        Attach metric modulations to generate score output:

        >>> staff = abjad.Staff("c'4 d'4 e'4 f'4 e'4 d'4")
        >>> score = abjad.Score([staff], name="Score")
        >>> abjad.attach(abjad.TimeSignature((3, 4)), staff[0])
        >>> metric_modulation = abjad.MetricModulation(
        ...     left_rhythm=abjad.Note("c4"),
        ...     right_rhythm=abjad.Note("c8."),
        ... )
        >>> abjad.attach(metric_modulation, staff[3], direction=abjad.UP)
        >>> abjad.override(staff).TextScript.staff_padding = 2.5
        >>> lilypond_file = abjad.LilyPondFile([r'\include "abjad.ily"', score])
        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(score)
            >>> print(string)
            \context Score = "Score"
            <<
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = 2.5
                }
                {
                    \time 3/4
                    c'4
                    d'4
                    e'4
                    f'4
                    ^ \markup \abjad-metric-modulation #2 #0 #3 #1 #'(1 . 1)
                    e'4
                    d'4
                }
            >>

    """

    left_rhythm: typing.Any
    right_rhythm: typing.Any
    hide: bool = False
    left_markup: _indicators.Markup | None = None
    right_markup: _indicators.Markup | None = None
    scale: tuple[int | float, int | float] = (1, 1)

    directed: typing.ClassVar[bool] = True

    def __post_init__(self):
        assert isinstance(self.hide, bool), repr(self.hide)
        if self.left_markup is not None:
            assert isinstance(self.left_markup, _indicators.Markup), repr(
                self.left_markup
            )
        if self.right_markup is not None:
            assert isinstance(self.right_markup, _indicators.Markup), repr(
                self.right_markup
            )
        assert isinstance(self.scale, tuple), repr(self.scale)

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is another metric modulation with the same
        ratio as this metric modulation.

        ..  container:: example

            >>> metric_modulation_1 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=abjad.Note("c'4."),
            ... )
            >>> metric_modulation_2 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ... )
            >>> notes = abjad.makers.make_notes([0], [(5, 16)])
            >>> metric_modulation_3 = abjad.MetricModulation(
            ...     left_rhythm=abjad.Note("c'4"),
            ...     right_rhythm=notes,
            ... )

            >>> metric_modulation_1.ratio
            (2, 3)
            >>> metric_modulation_2.ratio
            (2, 3)
            >>> metric_modulation_3.ratio
            (4, 5)

            >>> metric_modulation_1 == metric_modulation_1
            True
            >>> metric_modulation_1 == metric_modulation_2
            True
            >>> metric_modulation_1 == metric_modulation_3
            False

            >>> metric_modulation_2 == metric_modulation_1
            True
            >>> metric_modulation_2 == metric_modulation_2
            True
            >>> metric_modulation_2 == metric_modulation_3
            False

            >>> metric_modulation_3 == metric_modulation_1
            False
            >>> metric_modulation_3 == metric_modulation_2
            False
            >>> metric_modulation_3 == metric_modulation_3
            True

        """
        # custom definition because input rhythms don't compare:
        if isinstance(argument, type(self)):
            if self.ratio == argument.ratio:
                return True
        return False

    ### PRIVATE METHODS ###

    def _get_lilypond_command_string(self):
        if self._note_to_note():
            arguments = self._get_markup_arguments()
            left_exponent, left_dots, right_exponent, right_dots = arguments
            string = r"\abjad-metric-modulation"
            string += f" #{left_exponent} #{left_dots}"
            string += f" #{right_exponent} #{right_dots}"
        elif self._rhs_tuplet():
            arguments = self._get_markup_arguments()
            note_exponent, note_dots = arguments[:2]
            tuplet_exponent, tuplet_dots, tuplet_n, tuplet_d = arguments[2:]
            string = r"\abjad-metric-modulation-tuplet-rhs"
            string += f" #{note_exponent} #{note_dots}"
            string += f" #{tuplet_exponent} #{tuplet_dots}"
            string += f" #{tuplet_n} #{tuplet_d}"
        elif self._lhs_tuplet():
            arguments = self._get_markup_arguments()
            tuplet_exponent, tuplet_dots, tuplet_n, tuplet_d = arguments[:4]
            note_exponent, note_dots = arguments[4:]
            string = r"\abjad-metric-modulation-tuplet-lhs"
            string += f" #{tuplet_exponent} #{tuplet_dots}"
            string += f" #{tuplet_n} #{tuplet_d}"
            string += f" #{note_exponent} #{note_dots}"
        else:
            return None
        string += f" #'({self.scale[0]} . {self.scale[1]})"
        return string

    def _get_lilypond_format(self):
        markup = self._get_markup()
        return markup.string

    def _get_contributions(self, *, component=None, wrapper=None):
        contributions = _contributions.ContributionsBySite()
        if not self.hide:
            markup = self._get_markup()
            contributions = markup._get_contributions(
                component=component, wrapper=wrapper
            )
        return contributions

    def _get_markup(self):
        left_rhythm = self._initialize_rhythm(self.left_rhythm)
        right_rhythm = self._initialize_rhythm(self.right_rhythm)
        string = self._get_lilypond_command_string()
        if string is not None:
            markup = _indicators.Markup(rf"\markup {string}")
            return markup
        strings = []
        string = _illustrators.components_to_score_markup_string(left_rhythm)
        strings.extend(string.split("\n"))
        strings.append("=")
        strings.append(r"\hspace #-0.5")
        string = _illustrators.components_to_score_markup_string(right_rhythm)
        strings.extend(string.split("\n"))
        string = "\n".join(strings)
        string = rf"\markup {{ {string} }}"
        markup = _indicators.Markup(string)
        return markup

    # TODO: return namespace
    def _get_markup_arguments(self):
        left_rhythm = self._initialize_rhythm(self.left_rhythm)
        right_rhythm = self._initialize_rhythm(self.right_rhythm)
        if self._note_to_note():
            left_exponent = left_rhythm[0].written_duration.exponent
            left_dots = left_rhythm[0].written_duration.dot_count
            right_exponent = right_rhythm[0].written_duration.exponent
            right_dots = right_rhythm[0].written_duration.dot_count
            return (left_exponent, left_dots, right_exponent, right_dots)
        elif self._lhs_tuplet():
            tuplet_exponent = left_rhythm[0][0].written_duration.exponent
            tuplet_dots = left_rhythm[0][0].written_duration.dot_count
            tuplet_n, tuplet_d = left_rhythm[0].multiplier
            note_exponent = right_rhythm[0].written_duration.exponent
            note_dots = right_rhythm[0].written_duration.dot_count
            return (
                tuplet_exponent,
                tuplet_dots,
                tuplet_n,
                tuplet_d,
                note_exponent,
                note_dots,
            )
        elif self._rhs_tuplet():
            note_exponent = left_rhythm[0].written_duration.exponent
            note_dots = left_rhythm[0].written_duration.dot_count
            tuplet_exponent = right_rhythm[0][0].written_duration.exponent
            tuplet_dots = right_rhythm[0][0].written_duration.dot_count
            tuplet_n, tuplet_d = right_rhythm[0].multiplier
            return (
                note_exponent,
                note_dots,
                tuplet_exponent,
                tuplet_dots,
                tuplet_n,
                tuplet_d,
            )
        else:
            raise Exception("implement tied note values.")

    def _initialize_rhythm(self, rhythm):
        if isinstance(rhythm, _score.Component):
            selection = [rhythm]
        else:
            selection = rhythm
        return selection

    def _lhs_tuplet(self):
        left_rhythm = self._initialize_rhythm(self.left_rhythm)
        right_rhythm = self._initialize_rhythm(self.right_rhythm)
        if (
            isinstance(left_rhythm[0], _score.Tuplet)
            and len(left_rhythm[0]) == 1
            and isinstance(right_rhythm[0], _score.Note)
            and len(right_rhythm) == 1
        ):
            return True
        return False

    def _note_to_note(self):
        left_rhythm = self._initialize_rhythm(self.left_rhythm)
        right_rhythm = self._initialize_rhythm(self.right_rhythm)
        if (
            isinstance(left_rhythm[0], _score.Note)
            and len(left_rhythm) == 1
            and isinstance(right_rhythm[0], _score.Note)
            and len(right_rhythm) == 1
        ):
            return True
        return False

    def _rhs_tuplet(self):
        left_rhythm = self._initialize_rhythm(self.left_rhythm)
        right_rhythm = self._initialize_rhythm(self.right_rhythm)
        if (
            isinstance(left_rhythm[0], _score.Note)
            and len(left_rhythm) == 1
            and isinstance(right_rhythm[0], _score.Tuplet)
            and len(right_rhythm[0]) == 1
        ):
            return True
        return False

    @property
    def ratio(self) -> tuple[int, int]:
        """
        Gets ratio of metric modulation.

        ..  container:: example

            >>> metric_modulation = abjad.MetricModulation(
            ...     left_rhythm=abjad.Tuplet((2, 3), [abjad.Note("c'4")]),
            ...     right_rhythm=abjad.Note("c'4"),
            ... )
            >>> metric_modulation.ratio
            (2, 3)

        """
        left_duration = _getlib._get_duration(self.left_rhythm)
        right_duration = _getlib._get_duration(self.right_rhythm)
        duration = left_duration / right_duration
        pair = _duration.pair(duration)
        return pair



from .parser import LilyPondParser
from .reduced import parse_reduced_ly_syntax

_lilypond_parsers_by_language: dict = {}


def parse(string: str, language: str = "english", *, tag=None):
    r"""
    Parses LilyPond ``string``.

    ..  container:: example

        Parses LilyPond string with English note names:

        >>> container = abjad.parse("{c'4 d'4 e'4 f'4}")
        >>> abjad.show(container) # doctest: +SKIP

        Parses LilyPond string with Dutch note names:

        >>> container = abjad.parse("{c'8 des' e' fis'}", language='nederlands')
        >>> abjad.show(container) # doctest: +SKIP

        Tags output:

        >>> container = abjad.parse("{c'4 d'4 e'4 f'4}", tag=abjad.Tag("FOO"))
        >>> string = abjad.lilypond(container, tags=True)
        >>> print(string)
          %! FOO
        {
              %! FOO
            c'4
              %! FOO
            d'4
              %! FOO
            e'4
              %! FOO
            f'4
          %! FOO
        }

    Returns Abjad component.
    """
    if string.startswith("abj:"):
        return parse_reduced_ly_syntax(string[4:])
    if language not in _lilypond_parsers_by_language:
        parser = LilyPondParser(default_language=language)
        _lilypond_parsers_by_language[language] = parser
    parser = _lilypond_parsers_by_language[language]
    parser.tag = tag
    if hasattr(parser, "_guile"):
        parser._guile.tag = tag
    if hasattr(parser, "_lexdef"):
        parser._lexdef.tag = tag
    if hasattr(parser, "_syndef"):
        parser._syndef.tag = tag
    result = parser(string)
    return result



import copy

import pytest

import abjad


def test_Staff___copy___01():
    """
    Staves (shallow) copy grob overrides and context settings but not
    components.
    """

    staff_1 = abjad.Staff("c'8 d'8 e'8 f'8")
    abjad.override(staff_1).NoteHead.color = "#red"
    abjad.setting(staff_1).tupletFullLength = True

    staff_2 = copy.copy(staff_1)

    assert abjad.lilypond(staff_2) == abjad.string.normalize(
        r"""
        \new Staff
        \with
        {
            \override NoteHead.color = #red
            tupletFullLength = ##t
        }
        {
        }
        """
    )


def test_Staff___delitem___01():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    del staff[0]
    assert len(staff) == 4
    assert isinstance(staff[0], abjad.Rest)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Skip)
    assert isinstance(staff[3], abjad.Tuplet)
    del staff[0]
    assert len(staff) == 3
    assert isinstance(staff[0], abjad.Chord)
    assert isinstance(staff[1], abjad.Skip)
    assert isinstance(staff[2], abjad.Tuplet)
    del staff[0]
    assert len(staff) == 2
    assert isinstance(staff[0], abjad.Skip)
    assert isinstance(staff[1], abjad.Tuplet)
    del staff[0]
    assert len(staff) == 1
    assert isinstance(staff[0], abjad.Tuplet)
    del staff[0]
    assert len(staff) == 0


def test_Staff___delitem___02():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    del staff[-1]
    assert len(staff) == 4
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    del staff[-1]
    assert len(staff) == 3
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    del staff[-1]
    assert len(staff) == 2
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    del staff[-1]
    assert len(staff) == 1
    assert isinstance(staff[0], abjad.Note)
    del staff[-1]
    assert len(staff) == 0


def test_Staff___delitem___03():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    del staff[3]
    assert len(staff) == 4
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Tuplet)
    del staff[-2]
    assert len(staff) == 3
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Tuplet)
    del staff[2]
    assert len(staff) == 2
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    del staff[0]
    assert len(staff) == 1
    assert isinstance(staff[0], abjad.Rest)
    del staff[-1]
    assert len(staff) == 0


def test_Staff___getitem___01():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    assert isinstance(staff[-5], abjad.Note)
    assert isinstance(staff[-4], abjad.Rest)
    assert isinstance(staff[-3], abjad.Chord)
    assert isinstance(staff[-2], abjad.Skip)
    assert isinstance(staff[-1], abjad.Tuplet)


def test_Staff___getitem___02():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[0:0]
    assert len(components) == 0
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___03():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[0:1]
    assert len(components) == 1
    assert isinstance(components[0], abjad.Note)
    for x in staff:
        assert x._parent == staff
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___04():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[-1:]
    assert len(components) == 1
    assert isinstance(components[0], abjad.Tuplet)
    for x in components:
        assert x._parent == staff
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___05():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[1:-1]
    assert len(components) == 3
    assert isinstance(components[0], abjad.Rest)
    assert isinstance(components[1], abjad.Chord)
    assert isinstance(components[2], abjad.Skip)
    for x in components:
        assert x._parent == staff
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___06():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[2:]
    assert len(components) == 3
    assert isinstance(components[0], abjad.Chord)
    assert isinstance(components[1], abjad.Skip)
    assert isinstance(components[2], abjad.Tuplet)
    for x in components:
        assert x._parent == staff
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___07():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[:-2]
    assert len(components) == 3
    assert isinstance(components[0], abjad.Note)
    assert isinstance(components[1], abjad.Rest)
    assert isinstance(components[2], abjad.Chord)
    for x in components:
        assert x._parent == staff
    assert abjad.wf.wellformed(staff)


def test_Staff___getitem___08():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    components = staff[:]
    assert len(components) == 5
    assert isinstance(components[0], abjad.Note)
    assert isinstance(components[1], abjad.Rest)
    assert isinstance(components[2], abjad.Chord)
    assert isinstance(components[3], abjad.Skip)
    assert isinstance(components[4], abjad.Tuplet)
    assert all(_._parent is staff for _ in staff)


def test_Staff___init___01():
    """
    Initialize with context name.
    """

    staff = abjad.Staff(lilypond_type="BlueStaff")
    assert staff.lilypond_type == "BlueStaff"


def test_Staff___init___02():
    """
    Initialize with name.
    """

    staff = abjad.Staff(name="FirstBlueStaff")
    assert staff.name == "FirstBlueStaff"


def test_Staff___init___03():
    """
    Initialize with both context name and name.
    """

    staff = abjad.Staff(lilypond_type="BlueStaff", name="FirstBlueStaff")
    assert staff.lilypond_type == "BlueStaff"
    assert staff.name == "FirstBlueStaff"


def test_Staff___setitem___01():
    staff = abjad.Staff(
        [
            abjad.Note("c'4"),
            abjad.Rest((1, 4)),
            abjad.Chord([2, 3, 4], (1, 4)),
            abjad.Skip((1, 4)),
            abjad.Tuplet((4, 5), "c'16 c'16 c'16 c'16"),
        ]
    )

    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Rest)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    staff[1] = abjad.Chord([12, 13, 15], (1, 4))
    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Note)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    staff[0] = abjad.Rest((1, 4))
    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Rest)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Skip)
    assert isinstance(staff[4], abjad.Tuplet)
    staff[-2] = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Rest)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Tuplet)
    assert isinstance(staff[4], abjad.Tuplet)
    staff[-1] = abjad.Note(13, (1, 4))
    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Rest)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Chord)
    assert isinstance(staff[3], abjad.Tuplet)
    assert isinstance(staff[4], abjad.Note)
    staff[-3] = abjad.Skip((1, 4))
    assert len(staff) == 5
    assert abjad.wf.wellformed(staff)
    assert isinstance(staff[0], abjad.Rest)
    assert isinstance(staff[1], abjad.Chord)
    assert isinstance(staff[2], abjad.Skip)
    assert isinstance(staff[3], abjad.Tuplet)
    assert isinstance(staff[4], abjad.Note)


def test_Staff___setitem___02():
    """
    Reassign the entire contents of staff.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    assert staff._get_contents_duration() == abjad.Duration(4, 4)
    staff[:] = "c'8 c'8 c'8 c'8"
    assert staff._get_contents_duration() == abjad.Duration(4, 8)


def test_Staff___setitem___03():
    """
    Item-assign an empty container to staff.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    staff[0] = abjad.Voice([])


def test_Staff___setitem___04():
    """
    Slice-assign empty containers to staff.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    staff[0:2] = [abjad.Voice([]), abjad.Voice([])]


def test_Staff___setitem___05():
    """
    Bark when user assigns a slice to an item.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")

    with pytest.raises(AssertionError):
        staff[0] = [abjad.Note(2, (1, 4)), abjad.Note(2, (1, 4))]


def test_Staff___setitem___06():
    """
    Bark when user assigns an item to a slice.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")

    with pytest.raises(Exception):
        staff[0:2] = abjad.Note(2, (1, 4))


def test_Staff___setitem___07():
    """
    Slice-assign notes.
    """

    staff = abjad.Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    staff[0:4] = "d'8 d'8 d'8 d'8"
    assert len(staff) == 8
    for x in staff[0:4]:
        assert x.written_pitch == "d'"
    for x in staff[4:8]:
        assert x.written_pitch == "c'"
    assert abjad.wf.wellformed(staff)


def test_Staff___setitem___08():
    """
    Slice-assign chords.
    """

    staff = abjad.Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    chord = abjad.Chord([2, 3, 4], (1, 4))
    chords = abjad.mutate.copy(chord, 4)
    staff[0:4] = chords
    assert len(staff) == 8
    for x in staff[0:4]:
        assert x.written_duration == abjad.Duration(1, 4)
    for x in staff[4:8]:
        assert x.written_duration == abjad.Duration(1, 8)
    assert abjad.wf.wellformed(staff)


def test_Staff___setitem___09():
    """
    Slice-assign tuplets.
    """

    staff = abjad.Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    tuplet = abjad.Tuplet((2, 3), "c'8 c'8 c'8")
    tuplets = abjad.mutate.copy(tuplet, 2)
    staff[0:4] = tuplets
    assert len(staff) == 6
    for i, x in enumerate(staff):
        if i in [0, 1]:
            assert isinstance(x, abjad.Tuplet)
        else:
            assert isinstance(x, abjad.Note)
    assert abjad.wf.wellformed(staff)


def test_Staff_append_01():
    """
    Append one note.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    staff.append(abjad.Note("c'4"))
    assert abjad.wf.wellformed(staff)
    assert len(staff) == 5
    assert staff._get_contents_duration() == abjad.Duration(5, 4)


def test_Staff_append_02():
    """
    Append one chord.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    staff.append(abjad.Chord([2, 3, 4], (1, 4)))
    assert abjad.wf.wellformed(staff)
    assert len(staff) == 5
    assert staff._get_contents_duration() == abjad.Duration(5, 4)


def test_Staff_append_03():
    """
    Append one tuplet.
    """

    staff = abjad.Staff("c'4 c'4 c'4 c'4")
    staff.append(abjad.Tuplet((2, 3), "c'8 c'8 c'8"))
    assert abjad.wf.wellformed(staff)
    assert len(staff) == 5
    assert staff._get_contents_duration() == abjad.Duration(5, 4)


def test_Staff_engraver_consists_01():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff.consists_commands.append("Horizontal_bracket_engraver")
    staff.consists_commands.append("Instrument_name_engraver")

    assert abjad.wf.wellformed(staff)
    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        \with
        {
            \consists Horizontal_bracket_engraver
            \consists Instrument_name_engraver
        }
        {
            c'8
            d'8
            e'8
            f'8
        }
        """
    )


def test_Staff_engraver_removals_01():
    staff = abjad.Staff("c'8 d'8 e'8 f'8")
    staff.remove_commands.append("Time_signature_engraver")
    staff.remove_commands.append("Bar_number_engraver")

    assert abjad.lilypond(staff) == abjad.string.normalize(
        r"""
        \new Staff
        \with
        {
            \remove Time_signature_engraver
            \remove Bar_number_engraver
        }
        {
            c'8
            d'8
            e'8
            f'8
        }
        """
    )

    assert abjad.wf.wellformed(staff)



import pytest

import abjad


def test_dynamic_01():
    """
    Duplicate dynamics raise exception on attach.
    """

    voice = abjad.Voice("c'4")
    dynamic_1 = abjad.Dynamic("p")
    abjad.attach(dynamic_1, voice[0])
    dynamic_2 = abjad.Dynamic("p")
    with pytest.raises(Exception) as exception:
        abjad.attach(dynamic_2, voice[0])
    assert "attempting to attach conflicting indicator" in str(exception)



import platform

import pytest

import abjad


@pytest.mark.skipif(
    platform.python_implementation() != "CPython",
    reason="Benchmarking is only for CPython.",
)
def test_io_count_function_calls_01():
    result = abjad.io.count_function_calls("abjad.Note('c4')", global_context=globals())
    assert result < 4000


@pytest.mark.skipif(
    platform.python_implementation() != "CPython",
    reason="Benchmarking is only for CPython.",
)
def test_io_count_function_calls_02():
    result = abjad.io.count_function_calls(
        "abjad.Note(-12, (1, 4))", global_context=globals()
    )
    assert result < 400


def test_io_spawn_subprocess_01():
    assert abjad.io.spawn_subprocess('echo "hello world"') == 0
